@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2008-2013 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2008-2014 -- leonerd@leonerd.org.uk
 
 package Net::Async::HTTP;
 
@@ -9,7 +9,7 @@ use strict;
 use warnings;
 use base qw( IO::Async::Notifier );
 
-our $VERSION = '0.27_002';
+our $VERSION = '0.36_001';
 
 our $DEFAULT_UA = "Perl + " . __PACKAGE__ . "/$VERSION";
 our $DEFAULT_MAXREDIR = 3;
@@ -21,12 +21,16 @@ use Net::Async::HTTP::Connection;
 
 use HTTP::Request;
 use HTTP::Request::Common qw();
+use URI;
 
 use IO::Async::Stream 0.59;
 use IO::Async::Loop 0.59; # ->connect( handle ) ==> $stream
 
+use Future 0.28; # ->set_label
 use Future::Utils 0.16 qw( repeat );
 
+use Scalar::Util qw( blessed );
+use List::Util qw( first );
 use Socket qw( SOCK_STREAM IPPROTO_IP IP_TOS );
 BEGIN {
    if( $Socket::VERSION >= '2.010' ) {
@@ -50,6 +54,9 @@ use constant HTTPS_PORT => 443;
 use constant READ_LEN  => 64*1024; # 64 KiB
 use constant WRITE_LEN => 64*1024; # 64 KiB
 
+use Struct::Dumb;
+struct Ready => [qw( future connecting )];
+
 =head1 NAME
 
 C<Net::Async::HTTP> - use HTTP with C<IO::Async>
@@ -66,32 +73,23 @@ C<Net::Async::HTTP> - use HTTP with C<IO::Async>
 
  $loop->add( $http );
 
- $http->do_request(
+ my ( $response ) = $http->do_request(
     uri => URI->new( "http://www.cpan.org/" ),
+ )->get;
 
-    on_response => sub {
-       my ( $response ) = @_;
-       print "Front page of http://www.cpan.org/ is:\n";
-       print $response->as_string;
-       $loop->loop_stop;
-    },
-
-    on_error => sub {
-       my ( $message ) = @_;
-       print "Cannot fetch http://www.cpan.org/ - $message\n";
-       $loop->loop_stop;
-    },
- );
-
- $loop->loop_forever;
+ print "Front page of http://www.cpan.org/ is:\n";
+ print $response->as_string;
 
 =head1 DESCRIPTION
 
 This object class implements an asynchronous HTTP user agent. It sends
-requests to servers, and invokes continuation callbacks when responses are
-received. The object supports multiple concurrent connections to servers, and
-allows multiple requests in the pipeline to any one connection. Normally, only
-one such object will be needed per program to support any number of requests.
+requests to servers, returning L<Future> instances to yield responses when
+they are received. The object supports multiple concurrent connections to
+servers, and allows multiple requests in the pipeline to any one connection.
+Normally, only one such object will be needed per program to support any
+number of requests.
+
+As well as using futures the module also supports a callback-based interface.
 
 This module optionally supports SSL connections, if L<IO::Async::SSL> is
 installed. If so, SSL can be requested either by passing a URI with the
@@ -150,25 +148,25 @@ sub _init
    $self->{write_len} = WRITE_LEN;
 
    $self->{max_connections_per_host} = 1;
+
+   $self->{ssl_params} = {};
 }
 
 =head1 PARAMETERS
 
 The following named parameters may be passed to C<new> or C<configure>:
 
-=over 8
-
-=item user_agent => STRING
+=head2 user_agent => STRING
 
 A string to set in the C<User-Agent> HTTP header. If not supplied, one will
 be constructed that declares C<Net::Async::HTTP> and the version number.
 
-=item max_redirects => INT
+=head2 max_redirects => INT
 
 Optional. How many levels of redirection to follow. If not supplied, will
 default to 3. Give 0 to disable redirection entirely.
 
-=item max_in_flight => INT
+=head2 max_in_flight => INT
 
 Optional. The maximum number of in-flight requests to allow per host when
 pipelining is enabled and supported on that host. If more requests are made
@@ -176,81 +174,100 @@ over this limit they will be queued internally by the object and not sent to
 the server until responses are received. If not supplied, will default to 4.
 Give 0 to disable the limit entirely.
 
-=item max_connections_per_host => INT
+=head2 max_connections_per_host => INT
 
 Optional. Controls the maximum number of connections per hostname/server port
 pair, before requests will be queued awaiting one to be free. If not supplied,
 will default to 1. Give 0 to disable the limit entirely. See also the
 L</Connection Pooling> section documented above.
 
-=item timeout => NUM
+=head2 timeout => NUM
 
 Optional. How long in seconds to wait before giving up on a request. If not
 supplied then no default will be applied, and no timeout will take place.
 
-=item stall_timeout => NUM
+=head2 stall_timeout => NUM
 
 Optional. How long in seconds to wait after each write or read of data on a
 socket, before giving up on a request. This may be more useful than
 C<timeout> on large-file operations, as it will not time out provided that
 regular progress is still being made.
 
-=item proxy_host => STRING
+=head2 proxy_host => STRING
 
-=item proxy_port => INT
+=head2 proxy_port => INT
 
 Optional. Default values to apply to each C<request> method.
 
-=item cookie_jar => HTTP::Cookies
+=head2 cookie_jar => HTTP::Cookies
 
 Optional. A reference to a L<HTTP::Cookies> object. Will be used to set
 cookies in requests and store them from responses.
 
-=item pipeline => BOOL
+=head2 pipeline => BOOL
 
 Optional. If false, disables HTTP/1.1-style request pipelining.
 
-=item local_host => STRING
+=head2 local_host => STRING
 
-=item local_port => INT
+=head2 local_port => INT
 
-=item local_addrs => ARRAY
+=head2 local_addrs => ARRAY
 
-=item local_addr => HASH or ARRAY
+=head2 local_addr => HASH or ARRAY
 
 Optional. Parameters to pass on to the C<connect> method used to connect
 sockets to HTTP servers. Sets the local socket address to C<bind()> to. For
 more detail, see the documentation in L<IO::Async::Connector>.
 
-=item fail_on_error => BOOL
+=head2 fail_on_error => BOOL
 
 Optional. Affects the behaviour of response handling when a C<4xx> or C<5xx>
 response code is received. When false, these responses will be processed as
-other responses and passed to the C<on_response> callback, or used to set the
-successful result of the Future. When true, such an error response causes the
-C<on_error> handling or a failed Future instead. The HTTP response and request
-objects will be passed as well as the code and message.
+other responses and yielded as the result of the future, or passed to the
+C<on_response> callback. When true, such an error response causes the future
+to fail, or the C<on_error> callback to be invoked.
+
+The HTTP response and request objects will be passed as well as the code and
+message, and the failure name will be C<http>.
+
+ ( $code_message, "http", $response, $request ) = $f->failure
 
  $on_error->( "$code $message", $response, $request )
 
- ( $code_message, $response, $request ) = $f->failure
+=head2 read_len => INT
 
-=item read_len => INT
-
-=item write_len => INT
+=head2 write_len => INT
 
 Optional. Used to set the reading and writing buffer lengths on the underlying
 C<IO::Async::Stream> objects that represent connections to the server. If not
 define, a default of 64 KiB will be used.
 
-=item ip_tos => INT or STRING
+=head2 ip_tos => INT or STRING
 
 Optional. Used to set the C<IP_TOS> socket option on client sockets. If given,
 should either be a C<IPTOS_*> constant, or one of the string names
 C<lowdelay>, C<throughput>, C<reliability> or C<mincost>. If undefined or left
 absent, no option will be set.
 
-=back
+=head2 decode_content => BOOL
+
+Optional. If true, incoming responses that have a recognised
+C<Content-Encoding> are handled by the module, and decompressed content is
+passed to the body handling callback or returned in the C<HTTP::Response>. See
+L</CONTENT DECODING> below for details of which encoding methods are
+recognised. When this option is enabled, outgoing requests also have the
+C<Accept-Encoding> header added to them if it does not already exist.
+
+Currently the default is false, because this behaviour is new, but it may
+default to true in a later version. Applications which care which behaviour
+applies should set this to a defined value to ensure it doesn't change.
+
+=head2 SSL_*
+
+Additionally, any parameters whose names start with C<SSL_> will be stored and
+passed on requests to perform SSL requests. This simplifies configuration of
+common SSL parameters.
 
 =cut
 
@@ -262,11 +279,15 @@ sub configure
    foreach (qw( user_agent max_redirects max_in_flight max_connections_per_host
       timeout stall_timeout proxy_host proxy_port cookie_jar pipeline
       local_host local_port local_addrs local_addr fail_on_error
-      read_len write_len ))
+      read_len write_len decode_content ))
    {
       $self->{$_} = delete $params{$_} if exists $params{$_};
    }
 
+   foreach ( grep { m/^SSL_/ } keys %params ) {
+      $self->{ssl_params}{$_} = delete $params{$_};
+   }
+
    if( exists $params{ip_tos} ) {
       # TODO: This conversion should live in IO::Async somewhere
       my $ip_tos = delete $params{ip_tos};
@@ -287,6 +308,15 @@ sub configure
 
 =head1 METHODS
 
+The following methods documented with a trailing call to C<< ->get >> return
+L<Future> instances.
+
+When returning a Future, the following methods all indicate HTTP-level errors
+using the Future failure name of C<http>. If the error relates to a specific
+response it will be included. The original request is also included.
+
+ $f->fail( $message, "http", $response, $request )
+
 =cut
 
 sub connect_connection
@@ -308,7 +338,7 @@ sub connect_connection
       push @{ $args{extensions} }, "SSL";
    }
 
-   $conn->connect(
+   my $f = $conn->connect(
       host     => $host,
       service  => $port,
       ( map { defined $self->{$_} ? ( $_ => $self->{$_} ) : () } qw( local_host local_port local_addrs local_addr ) ),
@@ -316,11 +346,17 @@ sub connect_connection
       %args,
    )->on_done( sub {
       my ( $stream ) = @_;
+      $stream->configure(
+         notifier_name => "$host:$port,fd=" . $stream->read_handle->fileno,
+      );
+
       # Defend against ->setsockopt doing silly things like detecting SvPOK()
       $stream->read_handle->setsockopt( IPPROTO_IP, IP_TOS, $self->{ip_tos}+0 ) if defined $self->{ip_tos};
    })->on_fail( sub {
       $on_error->( $conn, "$host:$port - $_[0] failed [$_[-1]]" );
    });
+
+   $f->on_ready( sub { undef $f } ); # intentionally cycle
 }
 
 sub get_connection
@@ -337,61 +373,72 @@ sub get_connection
    my $conns = $self->{connections}{$key} ||= [];
    my $ready_queue = $self->{ready_queue}{$key} ||= [];
 
-   my $f = $args{future} || $self->loop->new_future;
-
    # Have a look to see if there are any idle connected ones first
    foreach my $conn ( @$conns ) {
-      $conn->is_idle and $conn->read_handle and return $f->done( $conn );
+      $conn->is_idle and $conn->read_handle and return Future->new->done( $conn );
    }
 
-   push @$ready_queue, $f unless $args{future};
+   my $ready = $args{ready};
+   $ready or push @$ready_queue, $ready =
+      Ready( $self->loop->new_future->set_label( "[connect $host:$port]" ), 0 );
 
-   if( !$self->{max_connections_per_host} or @$conns < $self->{max_connections_per_host} ) {
-      my $conn = Net::Async::HTTP::Connection->new(
-         notifier_name => "$host:$port",
-         max_in_flight => $self->{max_in_flight},
-         pipeline      => $self->{pipeline},
-         ready_queue   => $ready_queue,
-         read_len      => $self->{read_len},
-         write_len     => $self->{write_len},
+   my $f = $ready->future;
 
-         on_closed => sub {
-            my $conn = shift;
-
-            $conn->remove_from_parent;
-            @$conns = grep { $_ != $conn } @$conns;
-         },
-      );
-
-      $self->add_child( $conn );
-      push @$conns, $conn;
-
-      $self->connect_connection( %args,
-         conn => $conn,
-         on_error => sub {
-            my $conn = shift;
-
-            $f->fail( @_ );
-
-            @$conns = grep { $_ != $conn } @$conns;
-            @$ready_queue = grep { $_ != $f } @$ready_queue;
-
-            if( @$ready_queue ) {
-               # Requeue another connection attempt as there's still more to do
-               $self->get_connection( %args, future => $ready_queue->[0] );
-            }
-         },
-      );
+   my $max = $self->{max_connections_per_host};
+   if( $max and @$conns >= $max ) {
+      return $f;
    }
 
+   $ready->connecting++;
+
+   my $conn = Net::Async::HTTP::Connection->new(
+      notifier_name => "$host:$port,connecting",
+      ready_queue   => $ready_queue,
+      ( map { $_ => $self->{$_} }
+         qw( max_in_flight pipeline read_len write_len decode_content ) ),
+
+      on_closed => sub {
+         my $conn = shift;
+         my $http = $conn->parent;
+
+         $conn->remove_from_parent;
+         @$conns = grep { $_ != $conn } @$conns;
+
+         if( my $next = first { !$_->connecting } @$ready_queue ) {
+            # Requeue another connection attempt as there's still more to do
+            $http->get_connection( %args, ready => $next );
+         }
+      },
+   );
+
+   $self->add_child( $conn );
+   push @$conns, $conn;
+
+   $self->connect_connection( %args,
+      conn => $conn,
+      on_error => sub {
+         my $conn = shift;
+
+         $f->fail( @_ ) unless $f->is_cancelled;
+
+         @$conns = grep { $_ != $conn } @$conns;
+         @$ready_queue = grep { $_ != $ready } @$ready_queue;
+
+         if( my $next = first { !$_->connecting } @$ready_queue ) {
+            # Requeue another connection attempt as there's still more to do
+            $self->get_connection( %args, ready => $next );
+         }
+      },
+   );
+
    return $f;
 }
 
-=head2 $http->do_request( %args )
+=head2 $response = $http->do_request( %args )->get
 
-Send an HTTP request to a server, and set up the callbacks to receive a reply.
-The request may be represented by an L<HTTP::Request> object, or a L<URI>
-object, depending on the arguments passed.
+Send an HTTP request to a server, returning a L<Future> that will yield the
+response. The request may be represented by an L<HTTP::Request> object, or a
+L<URI> object, depending on the arguments passed.
 
 The following named arguments are used for C<HTTP::Request>s:
 
@@ -420,10 +467,10 @@ The following named arguments are used for C<URI> requests:
 
 =over 8
 
-=item uri => URI
+=item uri => URI or STRING
 
-A reference to a C<URI> object. If the scheme is C<https> then an SSL
-connection will be used.
+A reference to a C<URI> object, or a plain string giving the request URI. If
+the scheme is C<https> then an SSL connection will be used.
 
 =item method => STRING
 
@@ -431,37 +478,14 @@ Optional. The HTTP method. If missing, C<GET> is used.
 
 =item content => STRING or ARRAY ref
 
-Optional. The body content to use for C<POST> requests. If this is a plain
-scalar instead of an ARRAY ref, it will not be form encoded. In this case, a
-C<content_type> field must also be supplied to describe it.
-
-=item request_body => CODE or STRING
-
-Optional. Allows request body content to be generated by a callback, rather
-than being provided as part of the C<request> object. This can either be a
-C<CODE> reference to a generator function, or a plain string.
-
-As this is passed to the underlying L<IO::Async::Stream> C<write> method, the
-usual semantics apply here. If passed a C<CODE> reference, it will be called
-repeatedly whenever it's safe to write. The code should should return C<undef>
-to indicate completion.
-
-As with the C<content> parameter, the C<content_type> field should be
-specified explicitly in the request header, as should the content length
-(typically via the L<HTTP::Request> C<content_length> method). See also
-F<examples/PUT.pl>.
+Optional. The body content to use for C<PUT> or C<POST> requests. If this is a
+plain scalar instead of an ARRAY ref, it will not be form encoded. In this
+case, a C<content_type> field must also be supplied to describe it.
 
 =item content_type => STRING
 
 The type of non-form data C<content>.
 
-=item expect_continue => BOOL
-
-Optional. If true, sets the C<Expect> request header to the value
-C<100-continue> and does not send the C<request_body> parameter until a
-C<100 Continue> response is received from the server. If an error response is
-received then the C<request_body> code, if present, will not be invoked.
-
 =item user => STRING
 
 =item pass => STRING
@@ -477,7 +501,75 @@ Optional. Override the hostname or port number implied by the URI.
 
 =back
 
-For either request type, it takes the following continuation callbacks:
+For either request type, it takes the following arguments:
+
+=over 8
+
+=item request_body => STRING | CODE | Future
+
+Optional. Allows request body content to be generated by a future or
+callback, rather than being provided as part of the C<request> object. This
+can either be a plain string, a C<CODE> reference to a generator function, or
+a future.
+
+As this is passed to the underlying L<IO::Async::Stream> C<write> method, the
+usual semantics apply here. If passed a C<CODE> reference, it will be called
+repeatedly whenever it's safe to write. The code should should return C<undef>
+to indicate completion. If passed a C<Future> it is expected to eventually
+yield the body value.
+
+As with the C<content> parameter, the C<content_type> field should be
+specified explicitly in the request header, as should the content length
+(typically via the L<HTTP::Request> C<content_length> method). See also
+F<examples/PUT.pl>.
+
+=item expect_continue => BOOL
+
+Optional. If true, sets the C<Expect> request header to the value
+C<100-continue> and does not send the C<request_body> parameter until a
+C<100 Continue> response is received from the server. If an error response is
+received then the C<request_body> code, if present, will not be invoked.
+
+=item on_redirect => CODE
+
+Optional. A callback that is invoked if a redirect response is received,
+before the new location is fetched. It will be passed the response and the new
+URL.
+
+ $on_redirect->( $response, $location )
+
+=item on_body_write => CODE
+
+Optional. A callback that is invoked after each successful C<syswrite> of the
+body content. This may be used to implement an upload progress indicator or
+similar. It will be passed the total number of bytes of body content written
+so far (i.e. excluding bytes consumed in the header).
+
+ $on_body_write->( $written )
+
+=item max_redirects => INT
+
+Optional. How many levels of redirection to follow. If not supplied, will
+default to the value given in the constructor.
+
+=item timeout => NUM
+
+=item stall_timeout => NUM
+
+Optional. Overrides the object's configured timeout values for this one
+request. If not specified, will use the configured defaults.
+
+On a timeout, the returned future will fail with either C<timeout> or
+C<stall_timeout> as the operation name.
+
+ ( $message, "timeout" ) = $f->failure
+
+=back
+
+=head2 $http->do_request( %args )
+
+When not returning a future, the following extra arguments are used as
+callbacks instead:
 
 =over 8
 
@@ -494,12 +586,13 @@ server sent.
 Alternative to C<on_response>. A callback that is invoked when the header of a
 response has been received. It is expected to return a C<CODE> reference for
 handling chunks of body content. This C<CODE> reference will be invoked with
-no arguments once the end of the request has been reached.
+no arguments once the end of the request has been reached, and whatever it
+returns will be used as the result of the returned C<Future>, if there is one.
 
  $on_body_chunk = $on_header->( $header )
 
     $on_body_chunk->( $data )
-    $on_body_chunk->()
+    $response = $on_body_chunk->()
 
 =item on_error => CODE
 
@@ -514,34 +607,8 @@ well.
 
  $on_error->( $message, $response, $request )
 
-=item on_redirect => CODE
-
-Optional. A callback that is invoked if a redirect response is received,
-before the new location is fetched. It will be passed the response and the new
-URL.
-
- $on_redirect->( $response, $location )
-
-=item max_redirects => INT
-
-Optional. How many levels of redirection to follow. If not supplied, will
-default to the value given in the constructor.
-
-=item timeout => NUM
-
-=item stall_timeout => NUM
-
-Optional. Overrides the object's configured timeout values for this one
-request. If not specified, will use the configured defaults.
-
 =back
 
-=head2 $future = $http->do_request( %args )
-
-This method also returns a L<Future>, which will eventually yield the (final
-non-redirect) C<HTTP::Response>. If returning a future, then the
-C<on_response>, C<on_header> and C<on_error> callbacks are optional.
-
 =cut
 
 sub _do_one_request
@@ -561,6 +628,7 @@ sub _do_one_request
       host => $args{proxy_host} || $self->{proxy_host} || $host,
       port => $args{proxy_port} || $self->{proxy_port} || $port,
       SSL  => $args{SSL},
+      %{ $self->{ssl_params} },
       ( map { m/^SSL_/ ? ( $_ => $args{$_} ) : () } keys %args ),
    )->then( sub {
       my ( $conn ) = @_;
@@ -573,6 +641,18 @@ sub _do_one_request
    } );
 }
 
+sub _should_redirect
+{
+   my ( $response ) = @_;
+
+   # Should only redirect if we actually have a Location header
+   return 0 unless $response->is_redirect and defined $response->header( "Location" );
+
+   my $req_method = $response->request->method;
+   # Should only redirect GET or HEAD requests
+   return $req_method eq "GET" || $req_method eq "HEAD";
+}
+
 sub _do_request
 {
    my $self = shift;
@@ -608,17 +688,22 @@ sub _do_request
             $location = "http://$hostport" . $location;
          }
          else {
-            return $self->loop->new_future->fail( "Unrecognised Location: $location" );
+            return Future->new->fail( "Unrecognised Location: $location", http => $previous_response, $request );
          }
 
          my $loc_uri = URI->new( $location );
          unless( $loc_uri ) {
-            return $self->loop->new_future->fail( "Unable to parse '$location' as a URI" );
+            return Future->new->fail( "Unable to parse '$location' as a URI", http => $previous_response, $request );
          }
 
+         $self->debug_printf( "REDIRECT $loc_uri" );
+
          $args{on_redirect}->( $previous_response, $location ) if $args{on_redirect};
 
          %args = $self->_make_request_for_uri( $loc_uri, %args );
+         $request = $args{request};
+
+         undef $host; undef $port; undef $ssl;
       }
 
       my $uri = $request->uri;
@@ -653,23 +738,22 @@ sub _do_request
    while => sub {
       my $f = shift;
       return 0 if $f->failure or $f->is_cancelled;
-      return $response->is_redirect && $redirects--;
+      return _should_redirect( $response ) && $redirects--;
    } );
 
    if( $self->{fail_on_error} ) {
-      $future = $future->and_then( sub {
-         my $f = shift;
-         my $resp = $f->get;
+      $future = $future->then_with_f( sub {
+         my ( $f, $resp ) = @_;
          my $code = $resp->code;
 
          if( $code =~ m/^[45]/ ) {
             my $message = $resp->message;
             $message =~ s/\r$//; # HTTP::Message bug
 
-            return Future->new->fail( "$code $message", $resp, $request );
+            return Future->new->fail( "$code $message", http => $resp, $request );
          }
 
-         return $resp;
+         return $f;
       });
    }
 
@@ -684,6 +768,9 @@ sub do_request
    if( my $uri = delete $args{uri} ) {
       %args = $self->_make_request_for_uri( $uri, %args );
    }
+   elsif( !defined $args{request} ) {
+      croak "Require either 'uri' or 'request' argument";
+   }
 
    if( $args{on_header} ) {
       # ok
@@ -706,6 +793,7 @@ sub do_request
       croak "Expected 'on_response' or 'on_header' as CODE ref or to return a Future";
    }
 
+   my $on_error = delete $args{on_error};
    my $timeout = defined $args{timeout} ? $args{timeout} : $self->{timeout};
 
    my $future = $self->_do_request( %args );
@@ -714,7 +802,7 @@ sub do_request
       $future = Future->wait_any(
          $future,
          $self->loop->timeout_future( after => $timeout )
-                    ->transform( fail => sub { "Timed out" } ),
+                    ->transform( fail => sub { "Timed out", timeout => } ),
       );
    }
 
@@ -724,19 +812,19 @@ sub do_request
       $self->process_response( $response );
    } ) );
 
-   $future->on_fail( $args{on_error} ) if $args{on_error};
+   $future->on_fail( sub {
+      my ( $message, $name, @rest ) = @_;
+      $on_error->( $message, @rest );
+   }) if $on_error;
 
    # DODGY HACK:
    # In void context we'll lose reference on the ->wait_any Future, so the
    # timeout logic will never happen. So lets purposely create a cycle by
    # capturing the $future in on_done/on_fail closures within itself. This
    # conveniently clears them out to drop the ref when done.
-   if( !defined wantarray and $args{on_header} || $args{on_response} || $args{on_error} ) {
-      $future->on_done( sub { undef $future } );
-      $future->on_fail( sub { undef $future } );
-   }
+   return $future if defined wantarray;
 
-   return $future;
+   $future->on_ready( sub { undef $future } );
 }
 
 sub _make_request_for_uri
@@ -744,7 +832,12 @@ sub _make_request_for_uri
    my $self = shift;
    my ( $uri, %args ) = @_;
 
-   ref $uri and $uri->isa( "URI" ) or croak "Expected 'uri' as a URI reference";
+   if( !ref $uri ) {
+      $uri = URI->new( $uri );
+   }
+   elsif( blessed $uri and !$uri->isa( "URI" ) ) {
+      croak "Expected 'uri' as a URI reference";
+   }
 
    my $method = delete $args{method} || "GET";
 
@@ -758,13 +851,19 @@ sub _make_request_for_uri
 
       # Lack of content_type didn't used to be a failure condition:
       ref $args{content} or defined $args{content_type} or
-      carp "No 'content_type' was given with 'content'";
+         carp "No 'content_type' was given with 'content'";
 
       # This will automatically encode a form for us
       $request = HTTP::Request::Common::POST( $uri, Content => $args{content}, Content_Type => $args{content_type} );
    }
    else {
       $request = HTTP::Request->new( $method, $uri );
+      if( defined $args{content} ) {
+         defined $args{content_type} or carp "No 'content_type' was given with 'content'";
+
+         $request->content( $args{content} );
+         $request->content_type( $args{content_type} // "" );
+      }
    }
 
    $request->protocol( "HTTP/1.1" );
@@ -789,25 +888,49 @@ sub _make_request_for_uri
    return %args;
 }
 
-=head2 $future = $http->GET( $uri, %args )
+=head2 $response = $http->GET( $uri, %args )->get
 
-=head2 $future = $http->HEAD( $uri, %args )
+=head2 $response = $http->HEAD( $uri, %args )->get
 
-Convenient wrappers for using the C<GET> or C<HEAD> methods with a C<URI>
-object and few if any other arguments, returning a C<Future>.
+=head2 $response = $http->PUT( $uri, $content, %args )->get
+
+=head2 $response = $http->POST( $uri, $content, %args )->get
+
+Convenient wrappers for using the C<GET>, C<HEAD>, C<PUT> or C<POST> methods
+with a C<URI> object and few if any other arguments, returning a C<Future>.
+
+Remember that C<POST> with non-form data (as indicated by a plain scalar
+instead of an C<ARRAY> reference of form data name/value pairs) needs a
+C<content_type> key in C<%args>.
 
 =cut
 
 sub GET
 {
    my $self = shift;
-   return $self->do_request( method => "GET", uri => @_ );
+   my ( $uri, @args ) = @_;
+   return $self->do_request( method => "GET", uri => $uri, @args );
 }
 
 sub HEAD
 {
    my $self = shift;
-   return $self->do_request( method => "HEAD", uri => @_ );
+   my ( $uri, @args ) = @_;
+   return $self->do_request( method => "HEAD", uri => $uri, @args );
+}
+
+sub PUT
+{
+   my $self = shift;
+   my ( $uri, $content, @args ) = @_;
+   return $self->do_request( method => "PUT", uri => $uri, content => $content, @args );
+}
+
+sub POST
+{
+   my $self = shift;
+   my ( $uri, $content, @args ) = @_;
+   return $self->do_request( method => "POST", uri => $uri, content => $content, @args );
 }
 
 =head1 SUBCLASS METHODS
@@ -849,6 +972,123 @@ sub process_response
    $self->{cookie_jar}->extract_cookies( $response ) if $self->{cookie_jar};
 }
 
+=head1 CONTENT DECODING
+
+If the required decompression modules are installed and available, compressed
+content can be decoded. If the received C<Content-Encoding> is recognised and
+the required module is available, the content is transparently decoded and the
+decoded content is returned in the resulting response object, or passed to the
+data chunk handler. In this case, the original C<Content-Encoding> header will
+be deleted from the response, and its value will be available instead as
+C<X-Original-Content-Encoding>.
+
+The following content encoding types are recognised by these modules:
+
+=over 4
+
+=cut
+
+=item * gzip (q=0.7) and deflate (q=0.5)
+
+Recognised if L<Compress::Raw::Zlib> version 2.057 or newer is installed.
+
+=cut
+
+if( eval { require Compress::Raw::Zlib and $Compress::Raw::Zlib::VERSION >= 2.057 } ) {
+   my $make_zlib_decoder = sub {
+      my ( $bits ) = @_;
+      my $inflator = Compress::Raw::Zlib::Inflate->new(
+         -ConsumeInput => 0,
+         -WindowBits => $bits,
+      );
+      sub {
+         my $output;
+         my $status = @_ ? $inflator->inflate( $_[0], $output )
+                         : $inflator->inflate( "", $output, 1 );
+         die "$status\n" if $status && $status != Compress::Raw::Zlib::Z_STREAM_END();
+         return $output;
+      };
+   };
+
+   # RFC1950
+   __PACKAGE__->register_decoder(
+      deflate => 0.5, sub { $make_zlib_decoder->( 15 ) },
+   );
+
+   # RFC1952
+   __PACKAGE__->register_decoder(
+      gzip => 0.7, sub { $make_zlib_decoder->( Compress::Raw::Zlib::WANT_GZIP() ) },
+   );
+}
+
+=item * bzip2 (q=0.8)
+
+Recognised if L<Compress::Bzip2> version 2.10 or newer is installed.
+
+=cut
+
+if( eval { require Compress::Bzip2 and $Compress::Bzip2::VERSION >= 2.10 } ) {
+   __PACKAGE__->register_decoder(
+      bzip2 => 0.8, sub {
+         my $inflator = Compress::Bzip2::inflateInit();
+         sub {
+            return unless my ( $in ) = @_;
+            my $out = $inflator->bzinflate( \$in );
+            die $inflator->bzerror."\n" if !defined $out;
+            return $out;
+         };
+      }
+   );
+}
+
+=back
+
+Other content encoding types can be registered by calling the following method
+
+=head2 Net::Async::HTTP->register_decoder( $name, $q, $make_decoder )
+
+Registers an encoding type called C<$name>, at the quality value C<$q>. In
+order to decode this encoding type, C<$make_decoder> will be invoked with no
+paramters, and expected to return a CODE reference to perform one instance of
+decoding.
+
+ $decoder = $make_decoder->()
+
+This decoder will be invoked on string buffers to decode them until
+the end of stream is reached, when it will be invoked with no arguments.
+
+ $content = $decoder->( $encoded_content )
+ $content = $decoder->() # EOS
+
+=cut
+
+{
+   my %DECODERS; # {$name} = [$q, $make_decoder]
+
+   sub register_decoder
+   {
+      shift;
+      my ( $name, $q, $make_decoder ) = @_;
+
+      $DECODERS{$name} = [ $q, $make_decoder ];
+   }
+
+   sub can_decode
+   {
+      shift;
+      if( @_ ) {
+         my ( $name ) = @_;
+
+         return unless my $d = $DECODERS{$name};
+         return $d->[1]->();
+      }
+      else {
+         my @ds = sort { $DECODERS{$b}[0] <=> $DECODERS{$a}[0] } keys %DECODERS;
+         return join ", ", map { "$_;q=$DECODERS{$_}[0]" } @ds;
+      }
+   }
+}
+
 =head1 EXAMPLES
 
 =head2 Concurrent GET
@@ -891,18 +1131,26 @@ L<http://tools.ietf.org/html/rfc2616> - Hypertext Transfer Protocol -- HTTP/1.1
 
 =head1 SPONSORS
 
-Parts of this code were paid for by
+Parts of this code, or bugfixes to it were paid for by
 
 =over 2
 
 =item *
 
-Socialflow L<http://www.socialflow.com>
+SocialFlow L<http://www.socialflow.com>
 
 =item *
 
 Shadowcat Systems L<http://www.shadow.cat>
 
+=item *
+
+NET-A-PORTER L<http://www.net-a-porter.com>
+
+=item *
+
+Cisco L<http://www.cisco.com>
+
 =back
 
 =head1 AUTHOR
